from functools import wraps
import time
import os

import imageio
import open3d as o3d


def timeit(func):
    @wraps(func)
    def timeit_wrapper(*args, **kwargs):
        start_time = time.perf_counter()
        result = func(*args, **kwargs)
        end_time = time.perf_counter()
        total_time = end_time - start_time
        print(f'Function {func.__name__}{args} {kwargs} Took {total_time:.4f} seconds')
        return result

    return timeit_wrapper


def points2pcd(points):
    colors = [[0, 0, 1] for i in range(points.shape[0])]
    pcd = o3d.geometry.PointCloud()
    pcd.points = o3d.utility.Vector3dVector(points)
    pcd.colors = o3d.utility.Vector3dVector(colors)

    return pcd


def load_pcd(filename, normalize=False):
    pcd = o3d.io.read_point_cloud(filename)

    return pcd


def normalize_pcd(pcd):
    bbox = pcd.get_axis_aligned_bounding_box()
    scale = 1.6 / max((bbox.max_bound - bbox.min_bound))
    pcd = pcd.translate(-pcd.get_center())
    pcd_normalized = pcd.scale(scale, center=[0, 0, 0])

    return pcd_normalized


def display_inlier_outlier(cloud, ind):
    inlier_cloud = cloud.select_by_index(ind)
    outlier_cloud = cloud.select_by_index(ind, invert=True)

    print("Showing outliers (red) and inliers (gray): ")
    outlier_cloud.paint_uniform_color([1, 0, 0])
    inlier_cloud.paint_uniform_color([0, 0, 1])
    o3d.visualization.draw_geometries([inlier_cloud, outlier_cloud],
                                      zoom=0.3412,
                                      front=[0.4257, -0.2125, -0.8795],
                                      lookat=[2.6172, 2.0475, 1.532],
                                      up=[-0.0694, -0.9768, 0.2024])


def generate_gif(filenames, output_name):
    """
    Generates GIF from a list of image file paths

    :param filenames: a list of image file paths
    :param output_path: output folder to save gif
    :return:
    """
    images = []
    for filename in filenames:
        images.append(imageio.imread(filename))
    imageio.mimsave(os.apth.join(os.path.dirname(filenames[0]), '{}.gif'.format(output_name), images, format='GIF'))
